@@ -44,22 +44,12 @@ Shell::
     cat ~/.bashrc | pyline.py -n -r '^#(.*)' 'rgx and rgx.group()'
 
 
-**History**
-
-# updated 2012.11.17, Wes Turner
-# updated 2005.07.21, thanks to Jacob Oscarson
-# updated 2006.03.30, thanks to Mark Eichin
-
-see: `<http://code.activestate.com/recipes/437932-pyline-a-grep-like-sed-like-command-line-tool/>`_
-
-
 """
 
 import csv
 import json
 import logging
 import operator
-import sys
 
 from collections import namedtuple
 
@@ -129,7 +119,28 @@ class PylineResult(Result):
             unicode(odelim).join(str(x) for x in record))
 
 
-def pyline(_input,
+def _import_path_module():
+    Path = None
+    try:
+        from path import path as Path
+    except ImportError:
+        try:
+            from pathlib import Path
+            pass
+        except ImportError:
+            log.error("pip install pathlib (or path.py)")
+            Path = str  # os.exists, os
+            pass
+    return Path
+
+Path = _import_path_module()
+
+
+def get_path_module():
+    return Path
+
+
+def pyline(iterable,
            cmd=None,
            modules=[],
            regex=None,
@@ -142,7 +153,7 @@ def pyline(_input,
     Pyline: process an iterable
 
     Args:
-        _input (iterable): iterable of strings (e.g. sys.stdin or a file)
+        iterable (iterable): iterable of strings (e.g. sys.stdin or a file)
         cmd (str): python command string
         modules ([str]): list of modules to import
         regex (str): regex pattern to match (with groups)
@@ -179,22 +190,15 @@ def pyline(_input,
             # cmd = "rgx and rgx.groupdict()"
         else:
             cmd = "line"
+        if path_tools:
+            cmd = "p"
 
     Path = None
     if path_tools:
-        try:
-            from path import path as Path
-        except ImportError:
-            try:
-                from pathlib import Path
-                pass
-            except ImportError:
-                log.error("pip install pathlib (or path.py)")
-                Path = str
-                pass
+        Path = get_path_module()
 
     try:
-        log.debug("_cmd: %r" % cmd)
+        log.info("_cmd: %r" % cmd)
         codeobj = compile(cmd, 'command', 'eval')
     except Exception as e:
         e.message = "%s\ncmd: %s" % (e.message, cmd)
@@ -216,17 +220,16 @@ def pyline(_input,
     # from itertools import imap, repeat
     # j = lambda args: imap(str, izip_longest(args, repeat(odelim)))
 
-    for i, line in enumerate(_input):
+    for i, line in enumerate(iterable):
         l = line
         w = words = [w for w in line.strip().split(idelim)]
 
         p = path = None
         if path_tools and line.rstrip():
             try:
-                path = Path(line) or None
+                p = path = Path(line.strip()) or None
             except Exception as e:
                 log.exception(e)
-                path = None
                 pass
 
         rgx = _rgx and _rgx.match(line) or None
@@ -236,6 +239,7 @@ def pyline(_input,
             result = eval(codeobj, globals(), locals())  # ...
         except Exception as e:
             e.cmd = cmd
+            log.exception(repr(cmd))
             log.exception(e)
             raise
         yield PylineResult(i, result)
@@ -433,116 +437,93 @@ class ResultWriter_html(ResultWriter):
         self._output.write('</table>\n')
 
 
-TEST_INPUT = """
-Lines
-=======
-Of a file
----------
-With Text
+def get_option_parser():
+    import optparse
+    prs = optparse.OptionParser(
+        usage="%prog: [options] \"<command>\"",
+        epilog=EPILOG)
 
-And Without
+    prs.add_option('-f',
+                   dest='file',
+                   action='store',
+                   default='-',
+                   help="Input file (default: '-' for stdin)")
 
-http://localhost/path/to/file?query#fragment
+    prs.add_option('-o', '--output-file',
+                   dest='output',
+                   action='store',
+                   default='-',
+                   help="Output file (default: '-' for stdout)")
+    prs.add_option('-O', '--output-filetype',
+                   dest='output_filetype',
+                   action='store',
+                   default='txt',
+                   help="Output filetype <txt|csv|tsv|json> (default: txt)")
 
-"""
-import tempfile
-import os
-try:
-    import StringIO as io
-except ImportError:
-    import io
-import unittest
+    prs.add_option('-F', '--input-delim',
+                   dest='idelim',
+                   action='store',
+                   default=None,
+                   help=('Strings input field delimiter to split line'
+                         ' into ``words`` by'
+                         ' (default: None (whitespace)``'))
+    prs.add_option('-d', '--output-delim',
+                   dest='odelim',
+                   default="\t",
+                   help=('String output delimiter for lists and tuples'
+                         ' (default: \t (tab))``'))
 
+    prs.add_option('-m', '--modules',
+                   dest='modules',
+                   action='append',
+                   default=[],
+                   help='Module name to import (default: []) see -p and -r')
 
-class TestPyline(unittest.TestCase):
+    prs.add_option('-p', '--path-tools',
+                   dest='path_tools',
+                   action='store_true',
+                   help='Create path objects from each ``line``')
 
-    def setUp(self, *args):
-        (self._test_file_fd, self.TEST_FILE) = tempfile.mkstemp(text=True)
-        fd = self._test_file_fd
-        os.write(fd, TEST_INPUT)
-        os.write(fd, self.TEST_FILE)
+    prs.add_option('-r', '--regex',
+                   dest='regex',
+                   action='store',
+                   help='Regex to compile and match as ``rgx``')
+    prs.add_option('-R', '--regex-options',
+                   dest='regex_options',
+                   action='store',
+                   default='im',
+                   help='Regex options: I L M S X U (see ``pydoc re``)')
 
-        self.log = logging.getLogger(self.__class__.__name__)
-        self.log.setLevel(logging.DEBUG)
+    prs.add_option("-s", "--sort-asc",
+                   dest="sort_asc",
+                   action='store',
+                   help="Sort Ascending by field number")
+    prs.add_option("-S", "--sort-desc",
+                   dest="sort_desc",
+                   action='store',
+                   help="Reverse the sort order")
 
-    def tearDown(self):
-        os.close(self._test_file_fd)
-        os.remove(self.TEST_FILE)
+    prs.add_option('-n', '--number-lines',
+                   dest='number_lines',
+                   action='store_true',
+                   help='Print line numbers of matches')
 
-    def test_pyline__function(self):
-        PYLINE_TESTS = (
-            {"cmd": "line"},
-            {"cmd": "words"},
-            {"cmd": "sorted(words)"},
-            {"cmd": "w[:3]"},
-            {"regex":r"(.*)"},
-            {"regex":r"(.*)", "cmd": "rgx and rgx.groups()"},
-            {"regex":r"(.*)", "cmd": "rgx and rgx.groups() or '#'"},
+    prs.add_option('-i', '--ipython',
+                   dest='start_ipython',
+                   action='store_true',
+                   help='Start IPython with results')
 
-#             "cmd": "rgx and rgx.groups()"},
-        )
-        _test_output = sys.stdout
-        for test in PYLINE_TESTS:
-            for line in pyline(io.StringIO(TEST_INPUT), **test):
-                print(line, file=_test_output)
+    prs.add_option('-v', '--verbose',
+                   dest='verbose',
+                   action='store_true',)
+    prs.add_option('-q', '--quiet',
+                   dest='quiet',
+                   action='store_true',)
+    prs.add_option('-t', '--test',
+                   dest='run_tests',
+                   action='store_true',)
 
-    def test_pyline_cmdline(self):
-        CMDLINE_TESTS = (
-            tuple(),
-            ("line",),
-            ("l",),
-            ("l", "-n"),
-            ("l and l[:5]",),
-            ("words",),
-            ("w",),
-            ("w", "-n"),
-            ("w", '-O', 'csv'),
-            ("w", '-O', 'csv', '-n'),
-
-            ("w", '-O', 'csv', '-s', '0'),
-            ("w", '-O', 'csv', '-s', '1'),
-            ("w", '-O', 'csv', '-s', '1,2'),
-            ("w", '-O', 'csv', '-S', '1'),
-            ("w", '-O', 'csv', '-S', '1', '-n'),
-
-            ("w", '-O', 'json'),
-            ("w", '-O', 'json', '-n'),
-
-            ("w", '-O', 'tsv'),
-
-            ("w", '-O', 'html'),
-
-            ("len(words) > 2 and words",),
-
-            ('-r', '(.*with.*)'),
-            ('-r', '(.*with.*)',            '-R', 'i'),
-            ('-r', '(?P<line>.*with.*)'),
-            ('-r', '(?P<line>.*with.*)',    '-O', 'json'),
-            ('-r', '(.*with.*)', 'rgx and {"n":i, "match": rgx.groups()[0]}',
-             '-O', 'json'),
-            ("-r", '(.*with.*)', '_rgx.findall(line)',
-             '-O', 'json'),
-
-            ('-m',
-             'os',
-             'os.path.isfile(line) and (os.stat(line).st_size, line)'),
-            #
-            ("-p", "p and p.isfile() and (p.size, p, p.stat())")
-        )
-
-        TEST_ARGS = ('-f', __file__)  # self.TEST_FILE)
-
-        for argset in CMDLINE_TESTS:
-            _args = TEST_ARGS + argset
-            self.log.debug("main%s" % str(_args))
-            try:
-                output = main(*_args)
-                for n in output and output or []:
-                    self.log.debug(n)
-            except Exception as e:
-                self.log.error("cmd: %s" % repr(_args))
-                self.log.exception(e)
-                raise
+    return prs
 
 
 def get_sort_function(opts):  # (sort_asc, sort_desc)
@@ -577,86 +558,10 @@ def get_sort_function(opts):  # (sort_asc, sort_desc)
 
 
 def main(*args):
-    import optparse
     import logging
     import sys
 
-    prs = optparse.OptionParser(
-        usage="%prog: [options] \"<command>\"",
-        epilog=EPILOG)
-
-    prs.add_option('-f',
-                   dest='file',
-                   action='store',
-                   default='-',
-                   help="Input file (default: '-' for stdin)")
-    prs.add_option('-o', '--output-file',
-                   dest='output',
-                   action='store',
-                   default='-',
-                   help="Output file (default: '-' for stdout)")
-
-    prs.add_option('-O', '--output-filetype',
-                   dest='output_filetype',
-                   action='store',
-                   default='txt',
-                   help="Output filetype <txt|csv|tsv|json> (default: txt)")
-
-    prs.add_option('-F', '--input-delim',
-                   dest='idelim',
-                   action='store',
-                   default=None,
-                   help=('Strings input field delimiter to split line'
-                         ' into ``words`` by'
-                         ' (default: None (whitespace)``'))
-    prs.add_option('-d', '--output-delim',
-                   dest='odelim',
-                   default="\t",
-                   help=('String output delimiter for lists and tuples'
-                         ' (default: \t (tab))``'))
-
-    prs.add_option('-m', '--modules',
-                   dest='modules',
-                   action='append',
-                   default=[],
-                   help='Module name to import (default: []) see -p and -r')
-    prs.add_option('-p', '--path-tools',
-                   dest='path_tools',
-                   action='store_true',
-                   help='Create path objects from each ``line``')
-    prs.add_option('-r', '--regex',
-                   dest='regex',
-                   action='store',
-                   help='Regex to compile and match as ``rgx``')
-    prs.add_option('-R', '--regex-options',
-                   dest='regex_options',
-                   action='store',
-                   default='im',
-                   help='Regex options: I L M S X U (see ``pydoc re``)')
-
-    prs.add_option("-s", "--sort-asc",
-                   dest="sort_asc",
-                   action='store',
-                   help="Sort Ascending by field number")
-    prs.add_option("-S", "--sort-desc",
-                   dest="sort_desc",
-                   action='store',
-                   help="Reverse the sort order")
-
-    prs.add_option('-n', '--number-lines',
-                   dest='number_lines',
-                   action='store_true',
-                   help='Print line numbers of matches')
-
-    prs.add_option('-v', '--verbose',
-                   dest='verbose',
-                   action='store_true',)
-    prs.add_option('-q', '--quiet',
-                   dest='quiet',
-                   action='store_true',)
-    prs.add_option('-t', '--test',
-                   dest='run_tests',
-                   action='store_true',)
+    prs = get_option_parser()
 
     args = args and list(args) or sys.argv[1:]
     (opts, args) = prs.parse_args(args)
@@ -676,16 +581,6 @@ def main(*args):
 
     sortfunc = get_sort_function(opts)
 
-    if opts.file is '-':
-        opts._file = sys.stdin
-    else:
-        opts._file = open(opts.file, 'r')
-
-    if opts.output is '-':
-        opts._output = sys.stdout
-    else:
-        opts._output = open(opts.output, 'w')
-
     cmd = ' '.join(args)
     if not cmd.strip():
         if opts.regex:
@@ -704,32 +599,51 @@ def main(*args):
 
     opts.attrs = PylineResult._fields
 
-    writer, output_func = ResultWriter.get_writer(
-        opts._output,
-        filetype=opts.output_filetype,
-        number_lines=opts.number_lines,
-        attrs=opts.attrs)
-    writer.header()
+    try:
+        if opts.file is '-':
+            opts._file = sys.stdin
+        else:
+            opts._file = open(opts.file, 'r')
 
-    if not sortfunc:
-        for result in pyline(opts._file, **opts.__dict__):
-            if not result.result:
-                continue  # TODO
-            output_func(result)
-    else:
-        results = []
-        for result in pyline(opts._file, **opts.__dict__):
-            if not result.result:
-                continue
-            results.append(result)
-        for result in sortfunc(results):
-            output_func(result)
+        if opts.output is '-':
+            opts._output = sys.stdout
+        else:
+            opts._output = open(opts.output, 'w')
 
-    opts._file.close()
-    writer.footer()
+        writer, output_func = ResultWriter.get_writer(
+            opts._output,
+            filetype=opts.output_filetype,
+            number_lines=opts.number_lines,
+            attrs=opts.attrs)
+        writer.header()
+
+        # if not sorting, return a result iterator
+        if not sortfunc:
+            for result in pyline(opts._file, **opts.__dict__):
+                if not result.result:
+                    # skip result if not bool(result.result)
+                    continue  # TODO
+
+                output_func(result)
+        # if sorting, return the sorted list
+        else:
+            results = []
+            for result in pyline(opts._file, **opts.__dict__):
+                if not result.result:
+                    # skip result if not bool(result.result)
+                    continue
+                results.append(result)
+            for result in sortfunc(results):
+                output_func(result)
+
+        writer.footer()
+    finally:
+        if getattr(opts._file, 'fileno', int)() not in (0, 1, 2):
+            opts._file.close()
+
+        if opts.output != '-':
+            opts._output.close()
 
-    if opts.output is not '-':
-        opts._output.close()
 
 if __name__ == "__main__":
     main()
